require "grognard/version"
require "grognard/exceptions"
require "active_support/core_ext/string"

module Grognard
  class Guard
    attr_reader :can_dos, :current_abilities

    def initialize(*arr_roles, hash_roles: {})
      @can_dos = {}
      @current_abilities = nil

      insert_array_roles(arr_roles) ||
        insert_hash_roles(hash_roles)
    end

    def be_a(name)
      @current_abilities = @can_dos[name.to_sym] if role_defined! name
      self
    end

    def add_role(name, object)
      @can_dos[name.to_sym] = object if abilities_defined!(object)
    end

    def remove_role(name)
      @can_dos.delete name.to_sym
    end

    def <<(object)
      add_role(object.name.underscore, object)
      self
    end

    def can?(action, object, subject)
      @current_abilities.allowed_to(object, subject).include? action
    end

    def abilities_defined?(object)
      object.respond_to?(:allowed_to)
    end

    def abilities_defined!(object)
      abilities_defined?(object) ||
        raise_no_such_abilites
    end

    def role_defined?(name)
      @can_dos.key?(name.to_sym)
    end

    def role_defined!(name)
      role_defined?(name) || raise_no_such_role
    end

    def method_missing(meth, *args, &block)
      meth = meth.to_s
      case
      when meth.match(/^be_a_(\w+)/)
        be_a($1, *args, &block)
      when meth.match(/^can_be_a_(\w+)?/)
        can_be_a?($1, *args, &block)
      when meth.match(/^can_(\w+)?/)
        can?($1.to_sym, *args, &block)
      else
        super
      end
    end

    def respond_to?(meth, include_private = false)
      meth.to_s.start_with?("be_a_") ||
        meth.to_s.start_with?("can_") ||
        super
    end

    protected

    def raise_no_such_abilites
      fail AbilitiesNotDefinedError.new
    end

    def raise_no_such_role
      fail RoleNotDefinedError.new
    end

    private

    def insert_array_roles(roles)
      return unless roles.kind_of? Array
      roles.each do |role|
        add_role(role.name.underscore, role)
      end
    end

    def insert_hash_roles(roles)
      return unless roles.kind_of? Hash
      roles.each do |key, role|
        add_role(key, role)
      end
    end

    alias_method :can_be_a?, :role_defined?
  end
end